"""Module defining models."""

from opennmt.models.language_model import LanguageModel
from opennmt.models.transformer import Transformer
from opennmt.models.sequence_to_sequence import SequenceToSequence, EmbeddingsSharingLevel
from opennmt.models.sequence_tagger import SequenceTagger
from opennmt.models.sequence_classifier import SequenceClassifier
